#!/usr/bin/env -S TS_NODE_LOG_ERROR=true ts-node-script
"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const promises_1 = __importDefault(require("fs/promises"));
const path_1 = __importDefault(require("path"));
const util_1 = __importDefault(require("util"));
const ansi_colors_1 = __importDefault(require("ansi-colors"));
const inflected_1 = require("inflected");
if (process.env.NODE_ENV === "test") {
    ansi_colors_1.default.enabled = false;
}
const CLI = {
    default(commandHandler) {
        !process.argv[2] ? commandHandler() : null;
    },
    command(commandName, commandHandler) {
        if (Array.isArray(commandName)) {
            return commandName.includes(process.argv[2]) ? commandHandler() : null;
        }
        commandName === process.argv[2] ? commandHandler() : null;
    }
};
CLI.default(printCommands);
CLI.command(["help", "h"], printCommands);
CLI.command(["init", "new"], async () => {
    let memServerDirectory = await getMemServerDirectory();
    let boilerplateDirectory = `${__dirname}/../memserver-boilerplate`;
    if (!memServerDirectory) {
        memServerDirectory = "./memserver";
        await promises_1.default.mkdir(memServerDirectory);
    }
    try {
        await promises_1.default.access(`${memServerDirectory}/index.ts`);
    }
    catch (error) {
        const indexCode = await promises_1.default.readFile(`${boilerplateDirectory}/index.ts`);
        await promises_1.default.writeFile(`${memServerDirectory}/index.ts`, indexCode);
        console.log(ansi_colors_1.default.cyan("[Memserver CLI] /memserver/index.ts created"));
    }
    try {
        await promises_1.default.access(`${memServerDirectory}/routes.ts`);
    }
    catch (error) {
        const routesCode = await promises_1.default.readFile(`${boilerplateDirectory}/routes.ts`);
        await promises_1.default.writeFile(`${memServerDirectory}/routes.ts`, routesCode);
        console.log(ansi_colors_1.default.cyan("[Memserver CLI] /memserver/routes.ts created"));
    }
    try {
        await promises_1.default.access(`${memServerDirectory}/initializer.ts`);
    }
    catch (error) {
        const initializerCode = await promises_1.default.readFile(`${boilerplateDirectory}/initializer.ts`);
        await promises_1.default.writeFile(`${memServerDirectory}/initializer.ts`, initializerCode);
        console.log(ansi_colors_1.default.cyan("[Memserver CLI] /memserver/initializer.ts created"));
    }
    await createFixtureAndModelFoldersIfNeeded(memServerDirectory);
});
CLI.command(["generate", "g"], async () => {
    const memServerDirectory = await getMemServerDirectory();
    const generationType = process.argv[3];
    const modelName = process.argv[4] ? process.argv[4].toLocaleLowerCase() : null;
    if (!memServerDirectory) {
        return console.log(ansi_colors_1.default.red("[Memserver CLI] cannot find /memserver folder. Did you run $ memserver init ?"));
    }
    else if (!generationType) {
        return console.log(ansi_colors_1.default.red("[Memserver CLI] generate should be either $ memserver g model [modelName] or $ memserver g fixtures"));
    }
    else if (generationType === "model") {
        return await generateModel(modelName, memServerDirectory);
    }
    else if (generationType === "fixtures") {
        return await generateFixtures(modelName, memServerDirectory);
    }
    console.log(ansi_colors_1.default.red(`[Memserver CLI] $ memserver ${process.argv[2]} ${process.argv[3]} ${process.argv[4] ||
        ""} does not exists, available commands:`));
    return printCommands();
});
CLI.command(["console", "c"], async () => await openConsole());
CLI.command(["version", "v"], async () => {
    console.log(ansi_colors_1.default.cyan("[Memserver CLI]"), JSON.parse((await promises_1.default.readFile(`${__dirname}/../package.json`)).toString()).version);
});
async function printCommands() {
    const config = JSON.parse((await promises_1.default.readFile(`${__dirname}/../package.json`)).toString());
    const highlight = (text) => ansi_colors_1.default.bold.cyan(text);
    console.log(`${highlight("[Memserver CLI v" + config.version + "] Usage:")} memserver ${ansi_colors_1.default.yellow("<command (Default: help)>")}

memserver init | new                    # Sets up the initial memserver folder structure
memserver generate model ${ansi_colors_1.default.yellow("[ModelName]")}    # Generates the initial files for a MemServer Model ${ansi_colors_1.default.cyan('[alias: "memserver g model"]')}
memserver generate fixtures             # Outputs your initial MemServer state as pure javascript fixture files
memserver generate fixtures ${ansi_colors_1.default.yellow("[ModelName]")} # Outputs your initial MemServer state for certain model as pure javascript fixture
memserver console                       # Starts a MemServer console in node.js ${ansi_colors_1.default.cyan('[alias: "memserver c"]')}`);
}
async function generateModel(modelName, memServerDirectory) {
    if (!modelName) {
        return console.log(ansi_colors_1.default.red("[Memserver CLI] Please put a modelName to the memserver generate. Example: $ memserver generate model user"));
    }
    await createFixtureAndModelFoldersIfNeeded(memServerDirectory);
    const modelFileName = inflected_1.dasherize(inflected_1.singularize(modelName));
    const fixtureFileName = inflected_1.dasherize(inflected_1.pluralize(modelName));
    try {
        await promises_1.default.access(`${memServerDirectory}/models/${modelFileName}.ts`);
    }
    catch (error) {
        const classModelName = inflected_1.classify(modelName);
        await promises_1.default.writeFile(`${memServerDirectory}/models/${modelFileName}.ts`, `import Model from 'memserver/model';

export default class ${classModelName} extends Model {
  constructor() {
    super();
  }
}`);
        console.log(ansi_colors_1.default.cyan(`[Memserver CLI] /memserver/models/${modelFileName}.ts created`));
    }
    try {
        await promises_1.default.access(`${memServerDirectory}/fixtures/${fixtureFileName}.ts`);
    }
    catch (error) {
        await promises_1.default.writeFile(`${memServerDirectory}/fixtures/${fixtureFileName}.ts`, `export default [
];`);
        console.log(ansi_colors_1.default.cyan(`[Memserver CLI] /memserver/fixtures/${fixtureFileName}.ts created`));
    }
}
// TODO: this has to change MemServer.start() will change
async function generateFixtures(modelName, memServerDirectory) {
    const Model = (await Promise.resolve().then(() => __importStar(require("./model")))).default;
    const startMemserver = (await Promise.resolve().then(() => __importStar(require("./index")))).default;
    const Server = await startMemserver();
    const ModelDefinitions = await getModelDefinitions(memServerDirectory);
    const targetModels = modelName
        ? [inflected_1.classify(inflected_1.singularize(modelName))]
        : Object.keys(ModelDefinitions);
    await Promise.all(targetModels.map(async (Model) => {
        const sortedState = ModelDefinitions[Model].findAll().sort(sortFunction);
        const arrayOfRecords = util_1.default.inspect(sortedState, {
            depth: null,
            maxArrayLength: null
        });
        const targetFileName = inflected_1.pluralize(inflected_1.dasherize(inflected_1.underscore(Model)));
        const fileRelativePath = `/fixtures/${targetFileName}.ts`;
        const fileAbsolutePath = `${memServerDirectory}${fileRelativePath}`;
        try {
            await promises_1.default.access(fileAbsolutePath);
            const previousModels = (await Promise.resolve().then(() => __importStar(require(fileAbsolutePath)))).default;
            if (JSON.stringify(previousModels.sort(sortFunction)) === JSON.stringify(sortedState)) {
                return;
            }
        }
        catch (error) {
        }
        finally {
            await promises_1.default.writeFile(fileAbsolutePath, `export default ${arrayOfRecords};`);
            console.log(ansi_colors_1.default.yellow(`[MemServer] data written to ${fileRelativePath}`));
        }
    }));
}
async function createFixtureAndModelFoldersIfNeeded(memServerDirectory) {
    let boilerplateDirectory = `${__dirname}/../memserver-boilerplate`;
    try {
        await promises_1.default.access(`${memServerDirectory}/fixtures`);
    }
    catch (error) {
        await promises_1.default.mkdir(`${memServerDirectory}/fixtures`);
        await recursiveCopy(`${boilerplateDirectory}/fixtures`, `${memServerDirectory}/fixtures`);
        console.log(ansi_colors_1.default.cyan("[Memserver CLI] /memserver/fixtures folder created"));
    }
    try {
        await promises_1.default.access(`${memServerDirectory}/models`);
    }
    catch (error) {
        await promises_1.default.mkdir(`${memServerDirectory}/models`);
        await recursiveCopy(`${boilerplateDirectory}/models`, `${memServerDirectory}/models`);
        console.log(ansi_colors_1.default.cyan("[Memserver CLI] /memserver/models folder created"));
    }
}
async function openConsole() {
    if (process.cwd().includes("memserver")) {
        throw new Error(ansi_colors_1.default.red("[Memserver CLI] You are in the memserver directory, go to the root of your project to start memserver console."));
    }
    const MemServer = await (await Promise.resolve().then(() => __importStar(require("./index")))).default();
    const repl = (await Promise.resolve().then(() => __importStar(require("repl")))).default;
    console.log(ansi_colors_1.default.cyan("[Memserver CLI]"), "Started MemServer node.js console - check window.Memserver, window.MemServer and import/use your models ;)");
    repl.start("> ");
}
async function getMemServerDirectory() {
    const cwd = process.cwd();
    const folders = cwd.split("/");
    const memServerIndex = folders.findIndex((path) => path === "memserver");
    if (memServerIndex !== -1) {
        return folders.slice(0, memServerIndex + 1).join("/");
    }
    else if ((await promises_1.default.readdir(".")).includes("memserver")) {
        return `${cwd}/memserver`;
    }
}
function sortFunction(a, b) {
    if (a.id > b.id) {
        return 1;
    }
    else if (a.id < b.id) {
        return -1;
    }
    return 0;
}
async function getModelDefinitions(memServerDirectory) {
    const models = await promises_1.default.readdir(`${memServerDirectory}/models`);
    const modelDefinitions = await Promise.all(models.map(async (modelPath) => {
        return (await Promise.resolve().then(() => __importStar(require(`${memServerDirectory}/models/${modelPath}`)))).default;
    }));
    return modelDefinitions.reduce((result, modelDefinition) => {
        return Object.assign(result, { [modelDefinition.name]: modelDefinition });
    }, {});
}
async function recursiveCopy(sourcePath, targetPath) {
    try {
        await promises_1.default.access(sourcePath);
        let stats = await promises_1.default.stat(sourcePath);
        if (stats.isDirectory()) {
            try {
                await promises_1.default.access(targetPath);
            }
            catch {
                await promises_1.default.mkdir(targetPath);
            }
            let entries = await promises_1.default.readdir(sourcePath);
            await Promise.all(entries.map((entry) => recursiveCopy(path_1.default.join(sourcePath, entry), path_1.default.join(targetPath, entry))));
        }
        else {
            await promises_1.default.copyFile(sourcePath, targetPath);
        }
    }
    catch (error) {
        console.log('Recursive copy error:');
        throw error;
    }
}
