"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const ansi_colors_1 = __importDefault(require("ansi-colors"));
const fake_xml_http_request_1 = __importDefault(require("fake-xml-http-request"));
const model_1 = __importDefault(require("./model"));
const route_recognizer_1 = __importDefault(require("route-recognizer"));
require("pretender"); // NOTE: check this
require("./pretender-hacks"); // NOTE: check this
const DEFAULT_PASSTHROUGHS = [
    "http://localhost:0/chromecheckurl",
    "http://localhost:30820/socket.io"
];
class Memserver {
    constructor(options = { logging: true }) {
        this.Models = {};
        const initializer = options.initializer || async function () { };
        const routes = options.routes || function () { };
        const logging = options.hasOwnProperty("logging") ? options.logging : true;
        window.MemserverModel = window.MemserverModel || model_1.default;
        const initializerReturn = initializer();
        this.Models = window.MemserverModel._modelDefinitions;
        window.MemServer = startPretender(routes, Object.assign(options, { logging }), this.Models);
        window.MemServer.Models = this.Models;
        return window.MemServer;
    }
}
exports.default = Memserver;
function startPretender(routes, options, Models) {
    window.FakeXMLHttpRequest = fake_xml_http_request_1.default;
    window.RouteRecognizer = route_recognizer_1.default;
    window.Pretender.prototype.namespace = options.namespace;
    window.Pretender.prototype.urlPrefix = options.urlPrefix;
    window.Pretender.prototype.timing = options.timing;
    let pretender = new window.Pretender(function () {
        const Memserver = ansi_colors_1.default.cyan("[Memserver]");
        this.Models = Models;
        if (options.logging) {
            this.handledRequest = function (verb, path, request) {
                const method = verb.toUpperCase();
                console.log(Memserver, colorStatusCode(request.status), method, request.url);
                if (["POST", "PUT"].includes(method)) {
                    console.log(`${method} REQUEST BODY IS:`, request.params);
                }
                console.log(JSON.parse(request.responseText));
            };
            this.passthroughRequest = function (verb, path, request) {
                console.log(Memserver, ansi_colors_1.default.yellow("[PASSTHROUGH]"), verb, request.url);
            };
        }
        this.unhandledRequest = function (verb, path, request) {
            console.log(Memserver, ansi_colors_1.default.red("[UNHANDLED REQUEST]"), verb, path);
            console.log(ansi_colors_1.default.red("UNHANDLED REQUEST WAS:\n"), request);
            console.log(request);
        };
    }, { trackRequests: false });
    // HACK: Pretender this.passthrough for better UX
    // TODO: this doesnt passthrough full http:// https://
    pretender.passthrough = function (url) {
        const parent = window.Pretender.prototype;
        const verbs = ["get", "post", "put", "delete"];
        if (!url) {
            ["/**", "/", "/*"].forEach((path) => {
                verbs.forEach((verb) => pretender[verb](path, parent.passthrough));
            });
            return;
        }
        const fullUrl = (this.urlPrefix || "") + (this.namespace ? "/" + this.namespace : "") + url;
        verbs.forEach((verb) => pretender[verb](fullUrl, parent.passthrough));
    };
    DEFAULT_PASSTHROUGHS.forEach((url) => pretender.passthrough(url));
    // END: Pretender this.passthrough for better UX
    routes.apply(pretender, []);
    return pretender;
}
function colorStatusCode(statusCode) {
    if (statusCode === 200 || statusCode === 201) {
        return ansi_colors_1.default.green(statusCode);
    }
    else if (statusCode === 404 || statusCode === 204) {
        return ansi_colors_1.default.cyan(statusCode);
    }
    return ansi_colors_1.default.red(statusCode);
}
